%option 8bit
%option case-insensitive
%option nodefault
%option batch
%option never-interactive
%option stack
%option noyywrap
%option yylineno

%{

/*
 ******************************************************************************
 *
 * This file is a part of the IBIS ICM Parser.
 * Copyright (C) 2003 Kelly C. Green <kgreen22@mindspring.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 * 
 ******************************************************************************
 *
 * $Header: /home/kgreen/cvsroot/icmchk/src/ibis-icm-lex.l,v 1.2 2003/11/14 20:59:09 kgreen Exp $
 *
 ******************************************************************************
 *
 * $Log: ibis-icm-lex.l,v $
 * Revision 1.2  2003/11/14 20:59:09  kgreen
 * shift to v1.0.0
 *
 * Revision 1.1.1.1  2003/11/12 05:00:51  kgreen
 * Importing Project
 *
 *
 ******************************************************************************
 */

#if ! (defined(__GNUC__) || defined(__attribute__))
#define __attribute__(a) /* a */
#endif

static char *CVSID __attribute__((unused)) = "$Id: ibis-icm-lex.l,v 1.2 2003/11/14 20:59:09 kgreen Exp $";

%}


W ([\x20])
D ([0-9])
L ([A-Za-z])
A ([A-Za-z0-9])
S ([+\-])
C ([\!\"\#\$\%\&\'\(\)\*\,\:\;\<\>\?\@\\\^\`\{\|\}\~])
B ("["|"]")
X ([\.\/\=\_])
G ([\x21-\x7e])
P ([\x20-\x7e])

I ({S}?({D}+))
R ({S}?({D}*\.{D}+|{D}+\.{D}*))

E ([Ee]{S}?{I})
U ([TGMkmunpf]{L}*)
F (({I}|{R})({E}|{U}))

G_NO_C                   ({A}|{S}|{X}|{B})
P_NO_C                   ({W}|{G_NO_C})
P_NO_LB                  ({W}|{A}|{S}|{X}|{C}|"]")
G_NO_C_NO_LB             ({A}|{S}|{X}|"]")
P_NO_C_NO_LB             ({W}|{A}|{S}|{X}|"]")

P_FILENAME               ([a-z0-9_\-]+("."[a-z0-9_\-]{0,3})?)
P_NODE_NAME              (({A}|"_")+)


P_BEGIN_HEADER           ("[Begin"[ _]"Header]")
P_ICM_VER                ("[ICM"[ _]"Ver]")
P_FILE_NAME              ("[File"[ _]"Name]")
P_FILE_REV               ("[File"[ _]"Rev]")
P_DATE                   ("[Date]")
P_SOURCE                 ("[Source]")
P_NOTES                  ("[Notes]")
P_DISCLAIMER             ("[Disclaimer]")
P_COPYRIGHT              ("[Copyright]")
P_SUPPORT                ("[Support]")
P_REDISTRIBUTION         ("[Redistribution]")
P_REDISTRIBUTION_TEXT    ("[Redistribution"[ _]"Text]")
P_COMMENT_CHAR           ("[Comment"[ _]"Char]")
P_END_HEADER             ("[End"[ _]"Header]")
P_BEGIN_ICM_FAMILY       ("[Begin"[ _]"ICM"[ _]"Family]")
P_MANUFACTURER           ("[Manufacturer]")
P_ICM_FAMILY_DESCRIPTION ("[ICM"[ _]"Family"[ _]"Description]")
P_ICM_MODEL_LIST         ("[ICM"[ _]"Model"[ _]"List]")
P_BEGIN_ICM_MODEL        ("[Begin"[ _]"ICM"[ _]"Model]")
P_ICM_MODEL_DESCRIPTION  ("[ICM"[ _]"Model"[ _]"Description]")
P_TREE_PATH_DESCRIPTION  ("[Tree"[ _]"Path"[ _]"Description]")
P_NODAL_PATH_DESCRIPTION ("[Nodal"[ _]"Path"[ _]"Description]")
P_ICM_SWATH_PARAMETERS   ("[ICM"[ _]"Swath"[ _]"Parameters]")
P_ICM_SWATH_PIN_NUMBERS  ("[ICM"[ _]"Swath"[ _]"Pin"[ _]"Numbers]")
P_END_ICM_MODEL          ("[End"[ _]"ICM"[ _]"Model]")
P_ICM_PIN_MAP            ("[ICM"[ _]"Pin"[ _]"Map]")
P_ICM_NODE_MAP           ("[ICM"[ _]"Node"[ _]"Map]")
P_END_ICM_FAMILY         ("[End"[ _]"ICM"[ _]"Family]")
P_BEGIN_ICM_SECTION      ("[Begin"[ _]"ICM"[ _]"Section]")
P_DERIVATION_METHOD      ("[Derivation"[ _]"Method]")
P_RESISTANCE_MATRIX      ("[Resistance"[ _]"Matrix]")
P_INDUCTANCE_MATRIX      ("[Inductance"[ _]"Matrix]")
P_CONDUCTANCE_MATRIX     ("[Conductance"[ _]"Matrix]")
P_CAPACITANCE_MATRIX     ("[Capacitance"[ _]"Matrix]")
P_BANDWIDTH              ("[Bandwidth]")
P_ROW                    ("[Row]")
P_ICM_S_PARAMETER        ("[ICM"[ _]"S-parameter]")
P_END_ICM_SECTION        ("[End"[ _]"ICM"[ _]"Section]")
P_END                    ("[End]")

%x S_FUNDAMENTAL
%x S_SKIP_WS
%x S_FLUSH_LINE
%x S_FLUSH_KEYWORD
%x S_DISCARD_COMMENT
%x S_STRING
%x S_BLOCK
%x S_X_CHAR
%x S_NUMBER
%x S_WORD
%x S_NODE_NAME
%x S_FNAME
%x S_RATIO
%x S_FINAL

%{

/*
 * IBIS ICM v1.0 lexical analyzer.
 *
 * The purpose of the lexer is to break the input file up into
 * a series of tokens for the associated parser to process.
 *
 * Ideally, a lexer is largely stateless.  This one is not.
 * This is a consequence of the fact that the language being
 * parsed is not entirely regular.
 *
 * As a result, the lexical analyzer depends on the parser to give 
 * it hints, particularly with regard to distinguishing between 
 * numeric values, and names that may begin with or consist entirely 
 * of strings of digits.
 *
 */

#include "ibis-icm.h"
#include "ibis-icm-yacc.h"
#include "btree.h"
#include "dynstr.h"

/*
 * tab message control
 */

static int tab_count = 0;

/*
 * Macros to return input and/or lexer state and track 
 * of positional information and total length of line.
 *
 */

extern YYLTYPE yylloc;                     /* allocated by the parser */
static int yycolno = 1;                    /* maintain a column count */

#define IBIS_COLUMN_MAX 120

#define UNPUT(c) { if ((c) == '\n') { \
                     yycolno = 1; \
                   } else { \
                     yycolno -= 1; \
                   } \
                   unput((c)); \
                 }

#define UNPOP(c) { UNPUT((c)); yy_pop_state(); }

#define UNPUSH(c, a) { UNPUT((c)); yy_push_state((a)); }

/*
 * NOTE: the USER ACTION is executed before the action of any triggered rule
 */

static char *trash = (char *)0;          /* Set to a format string when trash is detected */

#define YY_USER_ACTION \
    yylloc.first_column = yycolno; \
    yycolno += yyleng; \
    yylloc.last_column  = yycolno  - 1; \
    if (yytext[yyleng-1] == '\n' && trash) { \
      err(&yylloc, trash); /* error is about previous line */ \
      trash = (char *)0; \
    } \
    yylloc.first_line = yylloc.last_line = yylineno; \
    if (yytext[yyleng-1] == '\n') { \
        if (1 < yylloc.first_column) yylloc.first_line = yylloc.last_line = yylineno - 1; \
        debug(&yylloc, "newline found at end of token."); \
        if (IBIS_COLUMN_MAX < (yylloc.last_column - 1)) { \
            err(&yylloc, "Line contains %i characters.  The maximum is %i.", \
                   yylloc.last_column - 1, IBIS_COLUMN_MAX); \
        } \
        yycolno = 1; \
    }

/*
 * utility macros, intended to make lexer actions more readable
 */

/*
 * don't screw up the line count when rejecting text containing a newline!!
 * (There's never more than one)
 * don't screw up the column count either.
 */

#define REJECT_CORRECTLY { \
                           if (strchr(yytext, '\n')) yylineno -= 1; \
                           if (yytext[yyleng-1] != '\n') yycolno -= yyleng; \
                           REJECT; \
                         }

/*
 * Comment characters are somewhat ugly, since they may be 
 * dynamically changed.  We have to provide a means of
 * identifying which of several possibilities is the actual one,
 * rejecting the inactive candidates for other rules to process.
 */

#define HANDLE_COMMENT  { if (comment_char == yytext[0]) { \
                            yy_push_state(S_DISCARD_COMMENT); \
                          } else { \
                            REJECT_CORRECTLY; \
                          } \
                        }

/*
 * small string helper functions.
 *
 * These routines are used to build up strings in dynamic memory.
 */

static char *string = (char *) 0;             /* used to hold temporary results */
YYLTYPE string_loc;                           /* used to capture location of string */
 
/*
 * Keyword and subparameter state maintained by the lexical analyzer,
 * with some assistance from the parser.
 */

static int keyw  = 0;                         /* current keyword id */
static int subp  = 0;                         /* current subparameter id */
static int enable_subp = 1;                   /* recognize keyword subparameters, parser may disable */

/*
 * A hint passed by the parser regarding the expected nature of the next token
 * This is a one shot, it is reset after use, successful or not.
 */

hint_t hint = H_DEFAULT;                /* PARSER HINT - take lexical analyzers default action */

/*
 * Input stream controls exerted by the parser.
 * Directs the lexical analyzer to reposition it's input.
 */

static int flush_keyword = 0;           /* PARSER CONTROL - position input to a keyword */
static int flush_line = 0;              /* PARSER CONTROL - advance input past next newline */
static char comment_char = '|';         /* PARSER CONTROL - the active current comment character */
static int enable_x_char = 0;           /* PARSER CONTROL - enable recognition of '[Comment char] X_char' */

/*
 * The lexical analyzer's symbol table.
 * All syntactic elements returned to the parser are stored here.
 *
 * Only a single unique copy of any particular string is ever stored.
 * It is never deleted, thus the pointer itself serves as a unique id.
 *
 * The parser may freely dereference, keep, compare 
 * or discard pointers without fear of compromising memory.
 * It MAY NOT free them however.
 */

static root_t string_table = (root_t)0; /* binary tree of ALL strings returned to the parser */

/*
 * Return various elements to the parser, 
 * while transparently maintaining lexer state.
 */

static int supress_newline = 0;          /* set whenever we need to supress the return of a newline token */
static int supress_tokens = 0;           /* set whenever we need supress the return of any and all tokens */

#define RETURN(c) { \
                    if ((c) == (unsigned int) '\n') { \
                       int repeated = supress_newline; \
                       supress_newline = 1; \
                       if (repeated) { /* return the first newline, but suppress any repeats */ \
                         debug(&yylloc, "lex newline: autosupressed a nil input line."); \
                       } else { \
                         if (supress_tokens) { \
                           debug(&yylloc, "lex token: supressed token <newline>."); \
                         } else { \
                           debug(&yylloc, "lex token: returned a newline."); \
                           return(END_OF_LINE); \
                         } \
                       } \
                    } else { \
                      supress_newline = 0; /* some intervening character, can't have repeated newline sequence now */ \
                      if (supress_tokens)  { \
                        debug(&yylloc, "lex token: supressed token %d.",(unsigned int) (c)); \
                      } else { \
                        return((c)); \
                      } \
                    } \
                  }

#define RETURN_STRING(s, type) { \
                                 if (s) { \
                                   yylval.str = yylex_string(s);  /* register string */ \
                                 } else { \
                                   yylval.str = yylex_string(""); /* return a constant empty string */ \
                                 } \
                                 debug(&yylloc, "lex token: type=%d, string \"%s\"", type, s); \
                                 RETURN((type)); \
                               }

/*
 * Keywords must begin in column one.  This requirement is explicitly tested here.
 * While is is possible to use the "^" syntax to bind keyword matches to column one,
 * the "^" pattern flag proved to be quite unreliable when the input stream is
 * manipulated in any way.  Since the input stream is manipulated extensively,
 * the "^" flags are not used.
 */

#define RETURN_KEYWORD(type, string) { \
                               if (1 == yylloc.first_column) {                                /* MUST start in column 1 */ \
                                 yylval.str = yylex_string(string);                           /* register string */ \
                                 keyw = type;                                                 /* retain keyword id */ \
                                 subp = 0;                                                    /* reset current subparameter */ \
                                 enable_subp = 1;                                             /* re-enable subparameters */ \
                                 debug(&yylloc, "lex token: type=%d, keyword \"%s\"", type, yytext); \
                                 if (END == type) BEGIN(S_FINAL);                             /* if END keyword, enter final state */ \
                                 RETURN((type));                                              /* return keyword */ \
                               } else { \
                                 warn(&yylloc, "Keyword %s, is NOT in column 1.", yytext);    /* warn now, parse error later */ \
                                 REJECT_CORRECTLY; \
                               } \
                             }

#define RETURN_SUBP_PARTIAL_HINT(key, type, string, hint) { \
                                                if ((keyw == key) && enable_subp) {     /* right keyword & subp is expectd */ \
                                                  if ('=' == yytext[yyleng-1]) yytext[yyleng-1] = ' '; /* suppress any trailing '=' */ \
                                                  subp = type;                          /* save new subparameter ID */ \
						  yycolno -= (yyleng - strlen(string)); /* fix column number */ \
                                                  yyless(strlen(string));               /* take only part of the token */ \
                                                  yylloc.last_column = yycolno - 1;     /* and location */ \
                                                  yylex_hint(hint);                     /* establish a hint, parser may reset */ \
                                                  debug(&yylloc, "lex token: subparameter \"%s\"", yytext); \
                                                  if (strncmp(yytext, string, strlen(string))) { \
						  err(&yylloc, \
							       "Subparameter '%s' is not correctly cased.  Use '%s'.", \
							       yytext, string); \
						  } \
                                                  RETURN_STRING(string, subp)           /* emit the subparameter */ \
						  } else { \
                                                  REJECT_CORRECTLY;                     /* it is not a valid subparameter */ \
                                                } \
                                              }

#define RETURN_SUBP_PARTIAL(key, type, string) RETURN_SUBP_PARTIAL_HINT((key), (type), (string), H_DEFAULT)

#define RETURN_SUBP(key, type, string) RETURN_SUBP_PARTIAL((key), (type), (string))

#define RETURN_SUBP_HINT(key, type, string, hint) RETURN_SUBP_PARTIAL_HINT((key), (type), (string), (hint))

/*
 * Return single characters, collapsing sequences of newline to just the first occurance.
 */

#define RETURN_CHAR(c) { \
                         if ((c) != (unsigned int) '\n') debug(&yylloc, "lex token: printable character, \"%c\"", (c)); \
                         RETURN((c)); \
                       }

%}

%%
    if (string) {                                                                       /* free any dangling temporary string */
      free(string);
      string = (char *)0;
    }

    if (flush_keyword) {                                                                /* advance input to keyword */
      flush_keyword = 0;
      debug(&yylloc, "lex input stream: flushing keyword");
      yy_push_state(S_FLUSH_KEYWORD);
    }

    if (flush_line) {                                                                   /* advance input to newline */
      flush_line = 0;
      debug(&yylloc, "lex input stream: flushing line");
      yy_push_state(S_FLUSH_LINE);
    }

    if (YY_START != INITIAL) {                                                          /* Consider the parser's hints only if parsing */
      switch (hint) {  /* take the parser's hint, if there is one */

        case H_STRING:    yy_push_state(S_STRING); 
                          string_loc.first_line = string_loc.last_line = yylloc.last_line; 
                          string_loc.first_column = string_loc.last_column = yylloc.last_column + 1; 
                          yy_push_state(S_SKIP_WS);                          debug(&yylloc, "lex state: String");        break;
        case H_BLOCK:     yy_push_state(S_BLOCK); 
                          string_loc.first_line = string_loc.last_line = yylloc.last_line; 
                          string_loc.first_column = string_loc.last_column = yylloc.last_column + 1; 
                                                                             debug(&yylloc, "lex state: Block");         break;
        case H_FNAME:     yy_push_state(S_FNAME);                            debug(&yylloc, "lex state: FileName");      break;
        case H_NUMBER:    yy_push_state(S_NUMBER);                           debug(&yylloc, "lex state: Number");        break;
        case H_NODE_NAME: yy_push_state(S_NODE_NAME);                        debug(&yylloc, "lex state: NodeName");      break;
        case H_SGR:       yy_push_state(S_RATIO);                            debug(&yylloc, "lex state: SGR Ratio");     break;
        default:          if (S_FUNDAMENTAL != yy_top_state())               debug(&yylloc, "lex state: Default (%d)", yy_top_state());
                          if (S_FUNDAMENTAL == yy_top_state())               debug(&yylloc, "lex state: Fundamental");   break;
      }

      hint = H_DEFAULT;                          /* reset the parser hint */
    }
   
<INITIAL>{

    {P_BEGIN_HEADER}                  { if (1 == yylloc.first_column) {                   /* must be in column 1 */
                                          yy_push_state(S_FUNDAMENTAL);                   /* begin parse */
                                          RETURN_KEYWORD(BEGIN_HEADER, "[Begin Header]");
                                        }
                                      }
}


<S_SKIP_WS>({W})?                     { string_loc.first_column = yylloc.last_column + 1; /* discard any leading white space */ }


<S_FLUSH_KEYWORD>{

    "["                               { debug(&yylloc, "FLUSH KEYWORD: possible resync at col %d char '%c' bol %d", 
                                              yylloc.first_column, yytext[0], YY_AT_BOL());
                                        if (1 == yylloc.first_column) {
                                          UNPOP(yytext[0]);
                                        } else {
                                          REJECT_CORRECTLY;
                                        }
                                      }
    {C}                               { HANDLE_COMMENT;                                               /* dispose of comments */ }
    {W}                               {                                                               /* discard white space */ }
    {G}                               { trash = "Orphaned line.";                                     /* detect residue */ }
    \n                                { supress_newline = 1; RETURN_CHAR(yytext[0]);                  /* suppress all newlines */ }
}

<S_FLUSH_LINE>{
    {C}                               { HANDLE_COMMENT;                                               /* dispose of comments */ }
    {W}                               {                                                               /* discard white space */ }
    {G}                               { trash = "Unexpected text precedes newline.";                  /* detect residue */ }
    \n                                { yy_pop_state(); RETURN_CHAR(yytext[0]);                       /* return newlines */ }
}

<S_DISCARD_COMMENT>{
    {P}                               { /* discard all printables up to newline */ }
    \n                                { UNPOP(yytext[0]); }
}

<S_FINAL>{
    {P}                               { /* Ignore all remaining printables */ }
    \n                                { RETURN_CHAR(yytext[0]);                             /* return only empty newlines */ }
}

<S_FUNDAMENTAL>{

    {P_BEGIN_HEADER}           { RETURN_KEYWORD(BEGIN_HEADER,           "[Begin Header]"); }
    {P_ICM_VER}                { RETURN_KEYWORD(ICM_VER,                "[ICM Ver]"); }
    {P_FILE_NAME}              { RETURN_KEYWORD(FILE_NAME,              "[File Name]"); }
    {P_FILE_REV}               { RETURN_KEYWORD(FILE_REV,               "[File Rev]"); }
    {P_DATE}                   { RETURN_KEYWORD(DATE,                   "[Date]"); }
    {P_SOURCE}                 { RETURN_KEYWORD(SOURCE,                 "[Source]"); }
    {P_NOTES}                  { RETURN_KEYWORD(NOTES,                  "[Notes]"); }
    {P_DISCLAIMER}             { RETURN_KEYWORD(DISCLAIMER,             "[Disclaimer]"); }
    {P_COPYRIGHT}              { RETURN_KEYWORD(COPYRIGHT,              "[Copyright]"); }
    {P_SUPPORT}                { RETURN_KEYWORD(SUPPORT,                "[Support]"); }
    {P_REDISTRIBUTION}         { RETURN_KEYWORD(REDISTRIBUTION,         "[Redistribution]"); }
    {P_REDISTRIBUTION_TEXT}    { RETURN_KEYWORD(REDISTRIBUTION_TEXT,    "[Redistribution Text]"); }
    {P_END_HEADER}             { RETURN_KEYWORD(END_HEADER,             "[End Header]"); }
    {P_BEGIN_ICM_FAMILY}       { RETURN_KEYWORD(BEGIN_ICM_FAMILY,       "[Begin ICM Family]"); }
    {P_MANUFACTURER}           { RETURN_KEYWORD(MANUFACTURER,           "[Manufacturer]"); }
    {P_ICM_FAMILY_DESCRIPTION} { RETURN_KEYWORD(ICM_FAMILY_DESCRIPTION, "[ICM Family Description]"); }
    {P_ICM_MODEL_LIST}         { RETURN_KEYWORD(ICM_MODEL_LIST,         "[ICM Model List]"); }
    {P_BEGIN_ICM_MODEL}        { RETURN_KEYWORD(BEGIN_ICM_MODEL,        "[Begin ICM Model]"); }
    {P_ICM_MODEL_DESCRIPTION}  { RETURN_KEYWORD(ICM_MODEL_DESCRIPTION,  "[ICM Model Description]"); }
    {P_TREE_PATH_DESCRIPTION}  { RETURN_KEYWORD(TREE_PATH_DESCRIPTION,  "[Tree Path Description]"); }
    {P_NODAL_PATH_DESCRIPTION} { RETURN_KEYWORD(NODAL_PATH_DESCRIPTION, "[Nodal Path Description]"); }
    {P_ICM_SWATH_PARAMETERS}   { RETURN_KEYWORD(ICM_SWATH_PARAMETERS,   "[ICM Swath Parameters]"); }
    {P_ICM_SWATH_PIN_NUMBERS}  { RETURN_KEYWORD(ICM_SWATH_PIN_NUMBERS,  "[ICM Swath Pin Numbers]"); }
    {P_END_ICM_MODEL}          { RETURN_KEYWORD(END_ICM_MODEL,          "[End ICM Model]"); }
    {P_ICM_PIN_MAP}            { RETURN_KEYWORD(ICM_PIN_MAP,            "[ICM Pin Map]"); }
    {P_ICM_NODE_MAP}           { RETURN_KEYWORD(ICM_NODE_MAP,           "[ICM Node Map]"); }
    {P_END_ICM_FAMILY}         { RETURN_KEYWORD(END_ICM_FAMILY,         "[End ICM Family]"); }
    {P_BEGIN_ICM_SECTION}      { RETURN_KEYWORD(BEGIN_ICM_SECTION,      "[Begin ICM Section]"); }
    {P_DERIVATION_METHOD}      { RETURN_KEYWORD(DERIVATION_METHOD,      "[Derivation Method]"); }
    {P_RESISTANCE_MATRIX}      { RETURN_KEYWORD(RESISTANCE_MATRIX,      "[Resistance Matrix]"); }
    {P_INDUCTANCE_MATRIX}      { RETURN_KEYWORD(INDUCTANCE_MATRIX,      "[Inductance Matrix]"); }
    {P_CONDUCTANCE_MATRIX}     { RETURN_KEYWORD(CONDUCTANCE_MATRIX,     "[Conductance Matrix]"); }
    {P_CAPACITANCE_MATRIX}     { RETURN_KEYWORD(CAPACITANCE_MATRIX,     "[Capacitance Matrix]"); }
    {P_BANDWIDTH}              { RETURN_KEYWORD(BANDWIDTH,              "[Bandwidth]"); }
    {P_ROW}                    { RETURN_KEYWORD(ROW,                    "[Row]"); }
    {P_ICM_S_PARAMETER}        { RETURN_KEYWORD(ICM_S_PARAMETER,        "[ICM S-parameter]"); }
    {P_END_ICM_SECTION}        { RETURN_KEYWORD(END_ICM_SECTION,        "[End ICM Section]"); }
    {P_END}                    { RETURN_KEYWORD(END,                    "[End]"); }

    {P_COMMENT_CHAR}           {  if (1 == yylloc.first_column) {                                /* handle this keyword directly */
                                    yy_push_state(S_FLUSH_KEYWORD);
                                    if (enable_x_char) {
                                      yy_push_state(S_X_CHAR);
                                    } else {
                                     err(&yylloc, "Use of the %s keyword is prohibited before the [ICM Ver] keyword.", yytext);
                                    }
                                  } else {
                                    warn(&yylloc, "Keyword %s, is NOT in column 1.", yytext);  /* warn now, error later */
                                    REJECT_CORRECTLY;
                                  }
                               }

    "["({A}|{W}|{S}|{X})+"]"   { if (1 == yylloc.first_column) { 
                                   err(&yylloc, "Unknown keyword '%s'.", yytext);
                                   yy_push_state(S_FLUSH_KEYWORD);
                                 } else {
                                   /* don't complain about unrecognized keywords not in column one */
                                    REJECT_CORRECTLY;
                                 } 
                               }

    "ICM_model_type"              { RETURN_SUBP(BEGIN_ICM_MODEL, ICM_model_type, "ICM_model_type"); }
    "SGR"                         { RETURN_SUBP_HINT(BEGIN_ICM_MODEL,SGR, "SGR", H_SGR); }
    "Ref_impedance"({W}|\t)*"="   { RETURN_SUBP_PARTIAL_HINT(BEGIN_ICM_MODEL, Ref_impedance_EQ, "Ref_impedance", H_NUMBER); }
    "Ref_impedance"({W}|\t)*      { err(&yylloc, "'%s' must be separated from its value by an '='.", "Ref_impedance");
                                    RETURN_SUBP_PARTIAL_HINT(BEGIN_ICM_MODEL,Ref_impedance_EQ, "Ref_impedance", H_NUMBER); }

    "Model_pinmap"                { RETURN_SUBP(TREE_PATH_DESCRIPTION, Model_pinmap, "Model_pinmap"); }
    "Fork"                        { RETURN_SUBP(TREE_PATH_DESCRIPTION, Fork, "Fork"); }
    "Endfork"                     { RETURN_SUBP(TREE_PATH_DESCRIPTION, Endfork, "Endfork"); }
    "Section"                     { RETURN_SUBP(TREE_PATH_DESCRIPTION, Section, "Section"); }

    "Model_nodemap"               { RETURN_SUBP(NODAL_PATH_DESCRIPTION, Model_nodemap, "Model_nodemap"); }
    "N_section"                   { RETURN_SUBP_HINT(NODAL_PATH_DESCRIPTION, N_section, "N_section", H_NODE_NAME); }

    "Left_edge"({W}|\t)*"="       { RETURN_SUBP_PARTIAL_HINT(ICM_SWATH_PARAMETERS, Left_edge_EQ, "Left_edge", H_NUMBER); }
    "Left_edge"({W}|\t)*          { err(&yylloc, "'%s' must be separated from its value by an '='.", "Left_edge");
                                    RETURN_SUBP_PARTIAL_HINT(ICM_SWATH_PARAMETERS, Left_edge_EQ, "Left_edge", H_NUMBER); }
    "Right_edge"({W}|\t)*"="      { RETURN_SUBP_PARTIAL_HINT(ICM_SWATH_PARAMETERS, Right_edge_EQ, "Right_edge", H_NUMBER); }
    "Right_edge"({W}|\t)*         { err(&yylloc, "'%s' must be separated from its value by an '='.", "Right_edge");
                                    RETURN_SUBP_PARTIAL_HINT(ICM_SWATH_PARAMETERS, Right_edge_EQ, "Right_edge", H_NUMBER); }
    "Top_edge"({W}|\t)*"="        { RETURN_SUBP_PARTIAL_HINT(ICM_SWATH_PARAMETERS, Top_edge_EQ, "Top_edge", H_NUMBER); }
    "Top_edge"({W}|\t)*           { err(&yylloc, "'%s' must be separated from its value by an '='.", "Top_edge");
                                    RETURN_SUBP_PARTIAL_HINT(ICM_SWATH_PARAMETERS, Top_edge_EQ, "Top_edge", H_NUMBER); }
    "Bottom_edge"({W}|\t)*"="     { RETURN_SUBP_PARTIAL_HINT(ICM_SWATH_PARAMETERS, Bottom_edge_EQ, "Bottom_edge", H_NUMBER); }
    "Bottom_edge"({W}|\t)*        { err(&yylloc, "'%s' must be separated from its value by an '='.", "Bottom_edge");
                                    RETURN_SUBP_PARTIAL_HINT(ICM_SWATH_PARAMETERS, Bottom_edge_EQ, "Bottom_edge", H_NUMBER); }

    "Pin_order"                   { RETURN_SUBP(ICM_PIN_MAP, Pin_order, "Pin_order"); }
    "Num_of_rows"({W}|\t)*"="     { RETURN_SUBP_PARTIAL_HINT(ICM_PIN_MAP, Num_of_rows_EQ, "Num_of_rows", H_NUMBER); }
    "Num_of_rows"({W}|\t)*        { err(&yylloc, "'%s' must be separated from its value by an '='.", "Num_of_rows");
                                    RETURN_SUBP_PARTIAL_HINT(ICM_PIN_MAP, Num_of_rows_EQ, "Num_of_rows", H_NUMBER); }
    "Num_of_columns"({W}|\t)*"="  { RETURN_SUBP_PARTIAL_HINT(ICM_PIN_MAP, Num_of_columns_EQ, "Num_of_columns", H_NUMBER); }
    "Num_of_columns"({W}|\t)*     { err(&yylloc, "'%s' must be separated from its value by an '='.", "Num_of_columns");
                                    RETURN_SUBP_PARTIAL_HINT(ICM_PIN_MAP, Num_of_columns_EQ, "Num_of_columns", H_NUMBER); }
    "Pin_list"                    { RETURN_SUBP(ICM_PIN_MAP, Pin_list, "Pin_list"); }

    "Mult"({W}|\t)*"="            { RETURN_SUBP_PARTIAL_HINT(TREE_PATH_DESCRIPTION, Mult_EQ, "Mult", H_NUMBER); }
    "Mult"({W}|\t)*               { err(&yylloc, "'%s' must be separated from its value by an '='.", "Mult");
                                    RETURN_SUBP_PARTIAL_HINT(TREE_PATH_DESCRIPTION, Mult_EQ, "Mult", H_NUMBER); }
    "Len"({W}|\t)*"="             { RETURN_SUBP_PARTIAL_HINT(TREE_PATH_DESCRIPTION, Len_EQ, "Len", H_NUMBER); }
    "Len"({W}|\t)*                { err(&yylloc, "'%s' must be separated from its value by an '='.", "Len");
                                    RETURN_SUBP_PARTIAL_HINT(TREE_PATH_DESCRIPTION, Len_EQ, "Len", H_NUMBER); }

    "Mult"({W}|\t)*"="            { RETURN_SUBP_PARTIAL_HINT(NODAL_PATH_DESCRIPTION, Mult_EQ, "Mult", H_NUMBER); }
    "Mult"({W}|\t)*               { err(&yylloc, "'%s' must be separated from its value by an '='.", "Mult");
                                    RETURN_SUBP_PARTIAL_HINT(NODAL_PATH_DESCRIPTION, Mult_EQ, "Mult", H_NUMBER); }
    "Len"({W}|\t)*"="             { RETURN_SUBP_PARTIAL_HINT(NODAL_PATH_DESCRIPTION, Len_EQ, "Len", H_NUMBER); }
    "Len"({W}|\t)*                { err(&yylloc, "'%s' must be separated from its value by an '='.", "Len");
                                    RETURN_SUBP_PARTIAL_HINT(NODAL_PATH_DESCRIPTION, Len_EQ, "Len", H_NUMBER); }

    "Banded_matrix"               { RETURN_SUBP(RESISTANCE_MATRIX, Banded_matrix, "Banded_matrix"); }
    "Diagonal_matrix"             { RETURN_SUBP(RESISTANCE_MATRIX, Diagonal_matrix, "Diagonal_matrix"); }
    "Full_matrix"                 { RETURN_SUBP(RESISTANCE_MATRIX, Full_matrix, "Full_matrix"); }
    "Sparse_matrix"               { RETURN_SUBP(RESISTANCE_MATRIX, Sparse_matrix, "Sparse_matrix"); }

    "Banded_matrix"               { RETURN_SUBP(INDUCTANCE_MATRIX, Banded_matrix, "Banded_matrix"); }
    "Diagonal_matrix"             { RETURN_SUBP(INDUCTANCE_MATRIX, Diagonal_matrix, "Diagonal_matrix"); }
    "Full_matrix"                 { RETURN_SUBP(INDUCTANCE_MATRIX, Full_matrix, "Full_matrix"); }
    "Sparse_matrix"               { RETURN_SUBP(INDUCTANCE_MATRIX, Sparse_matrix, "Sparse_matrix"); }

    "Banded_matrix"               { RETURN_SUBP(CONDUCTANCE_MATRIX, Banded_matrix, "Banded_matrix"); }
    "Diagonal_matrix"             { RETURN_SUBP(CONDUCTANCE_MATRIX, Diagonal_matrix, "Diagonal_matrix"); }
    "Full_matrix"                 { RETURN_SUBP(CONDUCTANCE_MATRIX, Full_matrix, "Full_matrix"); }
    "Sparse_matrix"               { RETURN_SUBP(CONDUCTANCE_MATRIX, Sparse_matrix, "Sparse_matrix"); }

    "Banded_matrix"               { RETURN_SUBP(CAPACITANCE_MATRIX, Banded_matrix, "Banded_matrix"); }
    "Diagonal_matrix"             { RETURN_SUBP(CAPACITANCE_MATRIX, Diagonal_matrix, "Diagonal_matrix"); }
    "Full_matrix"                 { RETURN_SUBP(CAPACITANCE_MATRIX, Full_matrix, "Full_matrix"); }
    "Sparse_matrix"               { RETURN_SUBP(CAPACITANCE_MATRIX, Sparse_matrix, "Sparse_matrix"); }

    "File_name"                   { RETURN_SUBP(ICM_S_PARAMETER, File_name, "File_name"); }
    "Port_assignment"             { RETURN_SUBP(ICM_S_PARAMETER, Port_assignment, "Port_assignment"); }

    {W}                           { /* consume leading white space, this is a terminator for WORDs */ }
    {G}                           { UNPUT(yytext[0]); 
                                    string_loc = yylloc; 
                                    yy_push_state(S_WORD);   /* DEFAULT: attempt to resolve a WORD */
                                  }

}

<S_WORD>{
    {C}                             { HANDLE_COMMENT;                                          /* Select active comment character */ }
    {C}                             { dyn_strcat(&string, yytext);                             /* select inactive comment character */ }
    ({G_NO_C})+                     { dyn_strcat(&string, yytext);                             /* add a glob of characters */ }
    {W}|\n                          { string_loc.last_column = yylloc.last_column - yyleng;    /* figure out string location */
                                      UNPOP(yytext[0]);                                        /* white space delineated */
				      yylloc = string_loc;                                     /* set location */
                                      if (string) RETURN_STRING(string, WORD);                 /* a WORD */
                                    }
}

<S_NODE_NAME>{
    {W}                             { /* discard leading white space  */ }
    {C}                             { HANDLE_COMMENT;                           /* Select active comment character */ }
    "("                             { RETURN_CHAR(yytext[0]);                   /* issue single '(', it starts a list of node names */ }
    ")"                             { yy_pop_state(); RETURN_CHAR(yytext[0]);   /* issue single ')', it ends a list of node names */ }
    "["                             { if (1 == yylloc.first_column) {
                                        UNPOP(yytext[0]);                       /* don't eat a keyword */
				      } else {
				        REJECT_CORRECTLY;                       /* Not keyword, let next rule handle it */
				      }
				    }
    {G}                             { UNPOP(yytext[0]); yy_push_state(WORD);    /* putback char, retry as a word */ }

    {P_NODE_NAME}/(")"|{W}|\t|\n)   { yy_pop_state(); RETURN_STRING(yytext, NODE_NAME);        /* return a NODE_NAME */ }

    {P_NODE_NAME}/{G}               { int len;
                                      string = dyn_strdup(yytext);
                                      for(len = strlen(string); len > 0; len--) UNPUT(string[len-1]);    /* put the string back */
                                      free(string);
                                      string = (char *) 0;
                                      yy_pop_state();
                                    } 
}

<S_X_CHAR>{
    {W}                             { /* discard leading white space  */ }
    {C}_char                        { char cchar = yylex_x_char(yytext[0]);
                                      yy_pop_state(); yy_push_state(S_FLUSH_LINE); supress_newline = 1;  /* consume the entire line */ 
                                      info(&yylloc, "Comment Character changed from '%c' to '%c'.", cchar, yytext[0]); }
    {G}_char                        { yy_pop_state(); yy_push_state(S_FLUSH_LINE); supress_newline = 1;  /* consume the entire line */ 
                                      err(&yylloc, "'%c_char' is not a permitted comment character.", yytext[0]); }
    _char                           { yy_pop_state(); yy_push_state(S_FLUSH_LINE); supress_newline = 1;  /* consume the entire line */ 
                                      err(&yylloc, "'_char' does not specify a comment character."); }

}

<S_RATIO>{
    {W}                             { /* discard leading white space  */ }
    {D}+:{D}+                       { yy_pop_state(); RETURN_STRING(yytext, SGR_RATIO); /* return an SGR Ratio */ }
    {I}({W}|\t)?:({W}|\t)?{I}       { err(&yylloc, "Malformed SGR ratio '%s'.  Use 'INT:INT' without embedded white space.", yytext);
                                      yy_pop_state(); RETURN_STRING(yytext, SGR_RATIO); /* return an SGR Ratio */ }
}

<S_FNAME>{
    {W}                             { /* discard leading white space  */ }
    {P_FILENAME}                    { if (strcspn(yytext, "ABCDEFGHIJKLMNOPQRSTUVWXYZ") == strlen(yytext)) {
                                        yy_pop_state(); RETURN_STRING(yytext, FILE_NAME);
                                      } else {
				        REJECT_CORRECTLY;  /* contains upper case characters, next pattern will match again */
                                      }
                                    }

    {G}+                            { err(&yylloc, "File Name contains illegal characters or is malformed.");
                                      yy_pop_state(); RETURN_STRING(yytext, FILE_NAME);        /* return bad file name */
                                    }
}

<S_NUMBER>{
    {W}                             { /* discard leading white space  */ }

    ({I}|{R}){E}                    { yy_pop_state(); RETURN_STRING(yytext, NUMBER);            /* standard scientific notation */ }

    ({I}|{R}){U}                    { int u = 0;                                                /* standard units notation */
                                      char unit;
                                      unit = yytext[u = strspn(yytext, "+-0123456789.")];
                                      debug(&yylloc, "lex numeric: numeric value \"%s\" has standard unit \"%c\"", yytext, yytext[u]);
                                      yytext[u] = 0;
                                      string = dyn_strdup(yytext);
                                      switch (unit) {
                                         case 'T': dyn_strcat(&string, "e+12"); break;
                                         case 'G': dyn_strcat(&string, "e+9"); break;
                                         case 'M': dyn_strcat(&string, "e+6"); break;
                                         case 'k': dyn_strcat(&string, "e+3"); break;
                                         case 'm': dyn_strcat(&string, "e-3"); break;
                                         case 'u': dyn_strcat(&string, "e-6"); break;
                                         case 'n': dyn_strcat(&string, "e-9"); break;
                                         case 'p': dyn_strcat(&string, "e-12"); break;
                                         case 'f': dyn_strcat(&string, "e-15"); break;
                                         default:  REJECT_CORRECTLY;
                                      }
                                      debug(&yylloc, "lex numeric: the conversion result is \"%s\"", string);
                                      yy_pop_state(); RETURN_STRING(string, NUMBER);
                                    }

    {R}                             { yy_pop_state(); RETURN_STRING(yytext, NUMBER);            /* real without an exponent or unit */ }

    {I}                             { yy_pop_state(); RETURN_STRING(yytext, NUMBER);            /* A simple integer is OK too... */ }
}

<S_STRING,S_BLOCK>{
    {C}                             { HANDLE_COMMENT; /* pick off active comment char */ }

    {C}                             { dyn_strcat(&string, yytext); /* include inactive comment characters */ }
    ({P_NO_C_NO_LB})+               { dyn_strcat(&string, yytext); /* everything except comment characters and "[" */ }
    \t                              { dyn_strcat(&string, yytext); /* capture tab in string/block */
                                      REJECT_CORRECTLY;            /* reject it too for tab counter */
                                    }
}

<S_STRING>(\n)                      { string_loc.last_column = yylloc.last_column - yyleng;
                                      UNPOP(yytext[0]);
                                      yylloc = string_loc;
                                      dyn_strrtrim(&string, ' '); 
                                      RETURN_STRING(string, TEXT_STRING); }

<S_BLOCK>{

    "["                             { if (1 == yylloc.first_column) {       /* check for a keyword... */
                                        string_loc.last_column = 1;
                                        dyn_strrtrim(&string, '\n');        /* trim trailing newlines (there's at least one) */
                                        UNPOP(yytext[0]); UNPUT('\n');      /* putback the last "[" and preceeding newline */
                                        yylloc = string_loc;
                                        RETURN_STRING(string, TEXT_BLOCK);
                                      } else {
                                        dyn_strcat(&string, yytext);        /* if not in column 1, "[" is nothing special */
                                      }
                                    }

    \n                              { string_loc.last_line += 1; dyn_strrtrim(&string, ' '); dyn_strcat(&string, yytext); } 
}

<*>{

    {W}   { /* skip spaces */ }

    {C}   { HANDLE_COMMENT; }

    \t    { if (!tab_count) warn(&yylloc, "One or more TAB characters were found. The use of TABs is discouraged.");
            tab_count += 1;  /* count and discard.  Anything that can actually use it (strings/blocks) already has */
          }

    \r    { yycolno -= 1; /* discard all return characters (a microsoft compatible feature :) */ }

    \n    {  /*
              * Handle newlines.  
              *
              * If we're in the INITIAL state, the newline (indeed most everything) is simply discarded.
              * Otherwise newlines are handled by FUNDAMENTAL state, or any state at all if they are supressed.
              * Otherwise put back the input and pop the state stack looking for any state that will accept it.
              */

             int state = YY_START;                                  /* WARNING, YY_START is a macro, snapshot it's value */

             if (INITIAL == state) {
               /* ignore the input */
             } else if (S_FUNDAMENTAL == state || supress_newline ) {
               RETURN_CHAR(yytext[0]);
             } else {
               UNPOP(yytext[0]);
             }
          }
              

    .     {  /*
              * If this an illegal char, we just gripe and pitch it.
              *
              * If we are in INITIAL state, everything is by definition junk, discard it.
              *
              * If we are in the FUNDAMENTAL state, then this is the ultimate catch-all rule.
              * We recognize comments and give everything else to the parser.
              *
              * Otherwise put back the input and pop the state stack in the hope that some previous 
              * state has a rule to match the character we're looking at right now.
              *
              */

             int state = YY_START;                  /* WARNING, YY_START is a macro, snapshot it's value */

             if ((yytext[0] < 0x20) || (yytext[0] > 0x7e)) {
               err(&yylloc, "Invalid character: <'%02.2x'> ignored.", (unsigned char)yytext[0]);
             } else if (INITIAL == state) {
                                                        /* ignore the input */
             } else if (S_FUNDAMENTAL == state) {       /* FUNDAMENTAL state */
               if (comment_char == yytext[0]) {         /* identify comments */
                 HANDLE_COMMENT;                        /* and push state to strip them out */
               } else {                                 /* otherwise */
                 RETURN_CHAR(yytext[0]);                /* return everything that gets here to the parser */
               }
             } else {                                   /* cannot handle this input in the current state */
               UNPOP(yytext[0]);                        /* pop the state stack and try again */
             }
          }

    <<EOF>> { yyterminate(); }
}

%%

/*
 * The lexical analyzer maintains a table of strings returned to
 * the parser.  This is because the parser depends on the continued
 * existence of these strings, which are passed to it by pointer.
 *
 * These strings are maintained in a balanced binary tree.
 * Only one copy of any given unique string is kept.
 * Strings may not be deleted until the entire parse is complete.
 *
 * What follows is the binary tree root, and some small routines
 * to manage it's elements, which the btree internally structures
 * as a collection of opaque objects of a uniform type 
 * referenced by void *.
 */

static int string_compare(void *item1, void *item2)
{
  return(strcmp((char *)item1, (char *)item2));
}

static void *string_create(void *item)
{
  return((void *)dyn_strdup((char *)item));
}

static void string_delete(void *item)
{
  free(item);
}

/*
 * Lexical analyzer's various public functions and controls
 */

void yylex_init (void)
{
  yycolno = 1;                 /* initial column count */
  supress_tokens = 0;          /* true to supress the return of all tokens */
  supress_newline = 0;         /* true if a newline should be supressed */
  trash = (char *)0;           /* set to a format string when trash characters are detected */

  comment_char = '|';          /* default comment char */

  keyw  = 0;                   /* current keyword id */
  subp  = 0;                   /* current subparameter id */
  enable_subp = 1;             /* enable recognition of subparameter reserved words */

  hint = H_DEFAULT;            /* PARSER HINT - take lexical analyzers default extion */

  flush_keyword = 0;           /* PARSER CONTROL - advance directly to the next keyword */
  flush_line = 0;              /* PARSER CONTROL - flush the remainder of an input line */

  if (string_table) btree_free(&string_table);          /* free any existing binary tree, deleting all members */

  string_table = btree_alloc(string_compare, string_create, string_delete);  /* allocate new binary tree */
  if (!string_table) {
    perror("fatal system error");
    abort();
    }

  if (string) {                                         /* free any temporary string, and start clean */
    free(string);
    string = (char *)0;
  }
}

/*
 * register and/or lookup a string in the lexical analyzer's string table.
 */

char *yylex_string (char *string)
{
return (char *) btree_search(string_table, string);
}

void yylex_flush_keyword (void)
{
  debug(&yylloc, "Advancing input stream to the next keyword");
  flush_keyword = 1;
}

void yylex_flush_line (void)
{
  debug(&yylloc, "Advancing input stream to the next newline");
  flush_line = 1;
}

void yylex_disable_subp (void)
{
  enable_subp = 0;
}

void yylex_hint (hint_t type)
{
  hint = type;
}

char yylex_x_char (char cchar)
{
  char c = comment_char;
  comment_char = cchar;
  return(c);
}

void yylex_enable_x_char (void)
{
  enable_x_char = 1;
  return;
}
