#ifndef BTREE_H
#define BTREE_H

/*
 ******************************************************************************
 *
 * This file is a part of the IBIS ICM Parser.
 * Copyright (C) 2003 Kelly C. Green <kgreen22@mindspring.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 * 
 ******************************************************************************
 *
 * $Header: /home/kgreen/cvsroot/icmchk/src/btree.h,v 1.1.1.1 2003/11/12 05:00:51 kgreen Exp $
 *
 ******************************************************************************
 *
 * $Log: btree.h,v $
 * Revision 1.1.1.1  2003/11/12 05:00:51  kgreen
 * Importing Project
 *
 *
 ******************************************************************************
 */

/*
 * include all necessary include files
 */

# include "includes.h"

/*
 * balanced binary tree utility routines
 */


#define TRUE (0==0)
#define FALSE (0==1)

typedef struct node_s node_t;

struct node_s {
  node_t *left;
  node_t *right;
  int balance;
  void *item;
} node_s;

typedef struct root_s *root_t;

struct root_s {
  node_t *node;
  int balanced;
  int depth;
  int (*compare)(void *item1, void *item2);
  void *(*create)(void *item);
  void (*delete)(void *item);
} root_s;

root_t btree_alloc PROTO((int   (*compare) PROTO((void *item1, void *item2)), 
                          void *(*create)  PROTO((void *item)), 
                          void  (*delete)  PROTO((void *item))));

int   btree_free   PROTO((root_t *root));

void *btree_find   PROTO((root_t root, void *item));

void *btree_search PROTO((root_t root, void *item));

void *btree_delete PROTO((root_t root, void *item));

int   btree_walk   PROTO((root_t root, 
                          void (*action) PROTO((void *item, void *data)), 
                          void *data));
#endif
