/*
 ******************************************************************************
 *
 * This file is a part of the IBIS ICM Parser.
 * Copyright (C) 2003 Kelly C. Green <kgreen22@mindspring.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 * 
 ******************************************************************************
 *
 * $Header: /home/kgreen/cvsroot/icmchk/src/dynstr.c,v 1.1.1.1 2003/11/12 05:00:51 kgreen Exp $
 *
 ******************************************************************************
 *
 * $Log: dynstr.c,v $
 * Revision 1.1.1.1  2003/11/12 05:00:51  kgreen
 * Importing Project
 *
 *
 ******************************************************************************
 */

#if ! (defined(__GNUC__) || defined(__attribute__))
#define __attribute__(a) /* a */
#endif

static char *RCSID __attribute__((unused)) = "$Id: dynstr.c,v 1.1.1.1 2003/11/12 05:00:51 kgreen Exp $";

#include "dynstr.h"

/*
 * Misc dynamic string utilities.
 *
 * These are intended to make it a bit easier/safer
 * to safely manipulate dynamically allocate strings.
 */
   
char *dyn_strcat (char **s1, char *s2)
{ /* second string (if any) is added to the first (replacing it, if any) */
    int l1 = 0;
    int l2 = 0;
    char *result;

    if (s1 && *s1) l1 = strlen(*s1);
    if (s2) l2 = strlen(s2);

    result = calloc(l1+l2+1,sizeof(char));
    if (!result) {
        perror("fatal system error");
        abort();
    }

    if (s1 && *s1 && l1) memcpy(&result[0], *s1, l1);
    if (s2 && l2) memcpy(&result[l1],s2,l2);

    if (s1 && *s1) free(*s1);
    if (s1) *s1 = result;

    return(result);
}

char *dyn_strdup (char *s1)
{ /* first string (if any) is duplicated */
    return(dyn_strcat((char **)0, s1));
}

char *dyn_strndup (char *s1, int n)
{
  char *temp = (char *) calloc(n + 1, sizeof(char));
  return strncpy(temp, s1, n);
}

char *dyn_strrtrim (char **s, char c)
{ /* zaps any trailing occurances of the given character */
    char *p;
    char *t;

    if (s && *s) {
        p = *s;
    } else {
        p = dyn_strdup("");
    }

    for (t = &p[strlen(p)]; (t != p) && (*--t == c); *t = '\0') ;
    if (s && *s) *s = p;

    return(p);
}

int dyn_strncasecmp (char *s1, char *s2, int n)
{
  char c1;
  char c2;

  while(n) {
    c1 = tolower((s1) ? *s1++ : 0);
    c2 = tolower((s2) ? *s2++ : 0);

    if (c1 < c2) return (-1);  /* different, s1 is 'smaller' */
    if (c1 > c2) return (+1);  /* different, s2 is 'larger' */
    if (0 == c1) return (0);   /* both strings are at end, 'equal' */
    n -= 1;
  }
  return (0);  /* no difference, 'equal' */
}

int dyn_strcasecmp (char *s1, char *s2)
{
  unsigned int len = 0;
  if (s1) len = strlen(s1);
  if (s2 && (len > strlen(s2))) len = strlen(s2);
  /*
   * use dyn_strncasecmp to do the actual check
   * Compare through the entire smaller string, 
   * including the terminating '\0' to ensure we 
   * return a proper value if the strings are identical 
   * through all of the first len charcters, but one 
   * of them is longer than the other (and therefore 
   * lexically 'greater than'.
   */
  return (dyn_strncasecmp(s1, s2, len+1));
}
