#ifndef IBIS_ICM_H
#define IBIS_ICM_H

/*
 ******************************************************************************
 *
 * This file is a part of the IBIS ICM Parser.
 * Copyright (C) 2003 Kelly C. Green <kgreen22@mindspring.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 * 
 ******************************************************************************
 *
 * $Header: /home/kgreen/cvsroot/icmchk/src/ibis-icm.h,v 1.3 2003/11/14 22:50:53 kgreen Exp $
 *
 ******************************************************************************
 *
 * $Log: ibis-icm.h,v $
 * Revision 1.3  2003/11/14 22:50:53  kgreen
 * address unix/microsoft %g differences
 *
 * Revision 1.2  2003/11/14 20:59:09  kgreen
 * shift to v1.0.0
 *
 * Revision 1.1.1.1  2003/11/12 05:00:51  kgreen
 * Importing Project
 *
 *
 ******************************************************************************
 */

/*
 * This is the IBIS ICM parser
 */

#define IBIS_ICM_VERSION "1.0.0"

/*
 * include all necessary include files
 */

# include "includes.h"

/*
 * Some platform dependent stuff
 * 
 * File pathname separator character
 *
 * "%g" format specifier
 * microsoft-C interprets the field width of %g differently from most everyone else
 * and this messes up some regression tests.
 */

#ifdef _WIN32
#define SEPARATOR "\\"
#define G_FMT "%#.5g"
#else
#define SEPARATOR "/"
#define G_FMT "%#g"
#endif

/*
 * Definitions for YACC/Bison generated code
 */

typedef struct yyltype
{
  int first_line;
  int first_column;

  int last_line;
  int last_column;
} yyltype;

#define YYLTYPE yyltype
#define YYERROR_VERBOSE

/*
 * Placeholder struct definitions for non-terminal parse types.
 * These are not of any interest to lexical analyzer, but are
 * required for the definition of YYSTYPE.
 */

typedef struct icm_s                       icm_t;
typedef struct icm_header_s                icm_header_t;
typedef struct icm_family_s                icm_family_t;
typedef struct icm_model_s                 icm_model_t;
typedef struct icm_tree_path_s             icm_tree_path_t;
typedef struct icm_node_path_s             icm_node_path_t;
typedef struct icm_swath_entry_s           icm_swath_entry_t;
typedef struct icm_pin_map_s               icm_pin_map_t;
typedef struct icm_pin_list_entry_s        icm_pin_list_entry_t;
typedef struct icm_node_map_s              icm_node_map_t;
typedef struct icm_node_list_entry_s       icm_node_list_entry_t;
typedef struct icm_section_s               icm_section_t;
typedef struct icm_port_assignment_entry_s icm_port_assignment_entry_t;
typedef struct icm_matrix_s                icm_matrix_t;
typedef struct icm_matrix_value_s          icm_matrix_value_t;


/*
 * Lexical analyzer "hint"s
 */

enum hint_e { H_DEFAULT,
              H_STRING,
              H_BLOCK,
              H_FNAME,
              H_WORD,
	      H_NODE_NAME,
              H_NUMBER,
              H_SGR
};

typedef enum hint_e hint_t;

/*
 * prototypes
 */

extern void yyparse_init          PROTO((void));
extern int  yyparse               PROTO((void));
extern void yyparse_finish        PROTO((void));

extern void yylex_init            PROTO((void));
extern int  yylex                 PROTO((void));
extern char *yylex_string         PROTO((char *string));
extern void yylex_flush_keyword   PROTO((void));
extern void yylex_flush_line      PROTO((void));
extern void yylex_disable_subp    PROTO((void));
extern void yylex_hint            PROTO((hint_t type));
extern char yylex_x_char          PROTO((char cchar));
extern void yylex_enable_x_char   PROTO((void));

extern void message_init          PROTO((int sort));
extern void message_finish        PROTO((void));
extern void set_verbosity         PROTO((int v));
extern int  get_verbosity         PROTO((void));
extern int  get_error_count       PROTO((void));
extern int  get_warning_count     PROTO((void));
extern int  get_info_count        PROTO((void));
extern int  get_debug_count       PROTO((void));
extern void err                   PROTO((YYLTYPE *loc, char *s, ...));
extern void warn                  PROTO((YYLTYPE *loc, char *s, ...));
extern void info                  PROTO((YYLTYPE *loc, char *s, ...));
extern void debug                 PROTO((YYLTYPE *loc, char *s, ...));

extern int file_exists            PROTO((char *name));
extern char *icm_file_name        PROTO((void));
extern char *icm_dir_name         PROTO((void));

#endif
